"""
MSeed object to hold and process MSeed data
"""
import os
from pathlib import Path
from typing import Dict, Tuple, List

from obspy.core import Stream

from sohstationviewer.controller.util import validate_file, validate_dir
from sohstationviewer.model.data_type_model import (
    DataTypeModel, ThreadStopped, ProcessingDataError)
from sohstationviewer.model.handling_data import read_mseed_or_text

from sohstationviewer.view.util.enums import LogType


class MSeed(DataTypeModel):
    """
    read and process mseed file into object with properties can be used to
    plot SOH data, mass position data, waveform data and gaps
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.nets_by_sta: Dict[str, List[str]] = {}
        self.processing_data()

    def finalize_data(self):
        """
        This function should be called after all folders finish reading to
            + get nets_by_sta from stream_header_by_key_chan
            + other tasks in super().finalize_data()

        """
        self.track_info("Retrieve nets.", LogType.INFO)
        self.nets_by_sta = self.retrieve_nets_from_stream_header(
            self.stream_header_by_key_chan)

        super().finalize_data()

    def read_folder(self, folder: str) -> Tuple[Dict]:
        """
        Read data streams for soh, mass position and waveform.
        For waveform, traces in data streams will be turn into memmap file
            to save resources
        :param folder: absolute path to data set folder
        :return waveform_data: waveform data by station
        :return soh_data: soh data by station
        :return mass_pos_data: mass position data by station
        :return gaps: gap list by station
        :return nets_by_sta: netcodes list by station
        """
        count = 0

        total = sum([len(files) for _, _, files in os.walk(folder)])
        for path, sub_dirs, files in os.walk(folder):
            try:
                validate_dir(path)
            except Exception as e:
                # skip Information folder
                self.track_info(str(e), LogType.WARNING)
                continue

            for file_name in files:
                if self.creator_thread.isInterruptionRequested():
                    raise ThreadStopped()

                path2file = Path(path).joinpath(file_name)
                if not validate_file(path2file, file_name):
                    continue
                count += 1
                if count % 10 == 0:
                    self.track_info(
                        f'Read {count} files/{total}', LogType.INFO)
                read_mseed_or_text(
                    path2file, self.tmp_dir, self.read_start, self.read_end,
                    self.stream_header_by_key_chan,
                    self.soh_data, self.mass_pos_data, self.waveform_data,
                    self.log_data, self.data_time,
                    self.req_soh_chans, self.req_wf_chans,
                    self.include_mp123zne, self.include_mp456uvw,
                    self.track_info, self.populate_cur_key_for_all_data)

        self.track_info(
            f'Skipped {total - count} invalid files.', LogType.INFO)

    def retrieve_nets_from_stream_header(
            self, streams: Dict[str, Dict[str, Stream]]) -> \
            Dict[str, List[str]]:
        """
        Retrive gaps by sta_id from stream_header_by_key_chan

        :param: dict of stream header by sta, chan
        :return nets_by_sta: list of nets by sta_id
        """
        nets_by_sta: Dict[str, List[str]] = {}
        for sta_id in streams:
            nets_by_sta[sta_id] = set()
            for chan_id in streams[sta_id]:
                stream = streams[sta_id][chan_id]
                nets = [tr.stats['network'].strip() for tr in stream]
                if len(nets) > 1:
                    for tr in stream:
                        tr.stats['network'] = stream[0].stats['network']
                nets_by_sta[sta_id].update(nets)
        return nets_by_sta

    def select_key(self) -> str:
        """
        :return selected_sta_id: the selected station id from available
            key of stream header.
            + If there is only one station id, return it.
            + If there is more than one, show all ids, let user choose one to
                return.
        """

        sta_ids = sorted(list(self.stream_header_by_key_chan.keys()))
        if len(sta_ids) == 0:
            msg = 'No required data found for the data set.'
            raise ProcessingDataError(msg)

        selected_sta_id = sta_ids[0]
        if len(sta_ids) > 1:
            msg = ("There are more than one stations in the given data.\n"
                   "Please select one to display")
            self.pause_signal.emit(msg, sta_ids)
            self.pause()
            selected_sta_id = sta_ids[self.pause_response]

        self.track_info(f'Select Station {selected_sta_id}', LogType.INFO)
        return selected_sta_id
