#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Suggested updates to obspy.io.reftek.core:
- modify section of the code that deals with upacking data
  with '16' and '32' encodings to account for number of samples in each data
  packet. Mass position channels data are '16' encoded.

Maeva Pourpoint IRIS/PASSCAL
"""

import copy
import obspy.io.reftek.core as obspy_rt130_core
import warnings

import numpy as np

from obspy import Trace, Stream, UTCDateTime
from obspy.core.util.obspy_types import ObsPyException
from obspy.io.reftek.packet import PACKET_FINAL_DTYPE
from obspy.io.reftek.util import _decode_ascii, _parse_long_time

from sohstationviewer.model.mseed_data.record_reader_helper import Unpacker
from sohstationviewer.model.reftek.from_rt2ms.packet import \
    (
    eh_et_payload_last_field_start, eh_et_payload_last_field_size, EHPacket,
)
from sohstationviewer.model.reftek.rt130_experiment.reftek_helper import \
    (
    read_rt130_file, convert_packet_to_obspy_format,
)


class DiscontinuousTrace(Trace):
    def __init__(self, *args, times, **kwargs):
        super().__init__(**kwargs)
        self.times = times



EH_PAYLOAD = {
    "station_name_extension": (35, 1, _decode_ascii),
    "station_name": (36, 4, _decode_ascii),
    "sampling_rate": (64, 4, float),
    "trigger_time": (72, 16, _parse_long_time),
    "first_sample_time": (
    eh_et_payload_last_field_start, eh_et_payload_last_field_size, _parse_long_time),
}


# class EHPacket(packet.EHPacket):
#     def __init__(self, data):
#         self._data = data
#         payload = self._data["payload"].tobytes()
#         for name, (start, length, converter) in EH_PAYLOAD.items():
#             data = payload[start:start + length]
#             if converter is not None:
#                 data = converter(data)
#             setattr(self, name, data)
#
#     def eh_et_info(self, nbr_DT_samples):
#         """
#         Compile EH and ET info to write to log file.
#         Returns list of strings.
#         Formatting of strings is based on earlier version of rt2ms.
#         """
#         info = []
#         # packet_tagline1 = ("\n\n{:s} exp {:02d} bytes {:04d} {:s} ID: {:s} "
#         #                    "seq {:04d}".format(self.type.decode(),
#         #                                        self.experiment_number,
#         #                                        self.byte_count,
#         #                                        Packet.time_tag(self.time),
#         #                                        self.unit_id.decode(),
#         #                                        self.packet_sequence))
#         # info.append(packet_tagline1)
#         # if self.type.decode('ASCII') == 'EH':
#         #     nbr_DT_samples = 0
#         #     info.append("\nEvent Header")
#         # else:
#         #     info.append("\nEvent Trailer")
#         # info.append("\n  event = " + str(self.event_number))
#         # info.append("\n  stream = " + str(self.data_stream_number + 1))
#         # info.append("\n  format = " + self.data_format.decode('ASCII'))
#         # info.append("\n  stream name = " + self.stream_name)
#         # info.append("\n  sample rate = " + str(self.sampling_rate))
#         # info.append("\n  trigger type = " + self.trigger_type)
#         trigger_time = Packet.time_tag(UTCDateTime(ns=self.trigger_time))
#         # info.append("\n  trigger time = " + trigger_time)
#         first_sample_time = Packet.time_tag(UTCDateTime(ns=self.first_sample_time))  # noqa: E501
#         # info.append("\n  first sample = " + first_sample_time)
#         # if self.last_sample_time:
#         #     info.append("\n  last sample = " + Packet.time_tag(UTCDateTime(ns=self.last_sample_time)))  # noqa: E501
#         # info.append("\n  bit weights = " + " ".join([val for val in self.channel_adjusted_nominal_bit_weights if val]))  # noqa: E501
#         # info.append("\n  true weights = " + " ".join([val for val in self.channel_true_bit_weights if val]))  # noqa: E501
#         packet_tagline2 = ("\nDAS: {:s} EV: {:04d} DS: {:d} FST = {:s} TT = "
#                            "{:s} NS: {:d} SPS: {:.1f} ETO: 0"
#                            .format(self.unit_id.decode(),
#                                    self.event_number,
#                                    self.data_stream_number + 1,
#                                    first_sample_time,
#                                    trigger_time,
#                                    nbr_DT_samples,
#                                    self.sampling_rate))
#         info.append(packet_tagline2)
#         return info


class Reftek130Exception(ObsPyException):
    pass


class Reftek130(obspy_rt130_core.Reftek130):
    @staticmethod
    def from_file(filename: str):
        rt130_unpacker = Unpacker('>')
        rt = Reftek130()
        rt._filename = filename
        packets_in_file = read_rt130_file(filename, rt130_unpacker)
        converted_packets = []
        for packet in packets_in_file:
            converted_packets.append(
                convert_packet_to_obspy_format(packet, rt130_unpacker))
        rt._data = np.array(converted_packets, dtype=PACKET_FINAL_DTYPE)
        return rt

    def to_stream(self, network="", location="", component_codes=None,
                  include_mp123=False, include_mp456=False,
                  headonly=False, verbose=False,
                  sort_permuted_package_sequence=False):
        """
        :type headonly: bool
        :param headonly: Determines whether or not to unpack the data or just
            read the headers.
        """
        if not len(self._data):
            msg = "No packet data in Reftek130 object (file: {})"
            raise Reftek130Exception(msg.format(self._filename))
        self.check_packet_sequence_and_sort(sort_permuted_package_sequence)
        self.check_packet_sequence_contiguous()
        self.drop_not_implemented_packet_types()
        if not len(self._data):
            msg = ("No packet data left in Reftek130 object after dropping "
                   "non-implemented packets (file: {})").format(self._filename)
            raise Reftek130Exception(msg)
        st = Stream()
        for event_number in np.unique(self._data['event_number']):
            data = self._data[self._data['event_number'] == event_number]
            # we should have exactly one EH and one ET packet, truncated data
            # sometimes misses the header or trailer packet.
            eh_packets = data[data['packet_type'] == b"EH"]
            et_packets = data[data['packet_type'] == b"ET"]
            if len(eh_packets) == 0 and len(et_packets) == 0:
                msg = ("Reftek data (file: {}) contain data packets without "
                       "corresponding header or trailer packet."
                       .format(self._filename))
                raise Reftek130Exception(msg)
            if len(eh_packets) > 1 or len(et_packets) > 1:
                msg = ("Reftek data (file: {}) contain data packets with "
                       "multiple corresponding header or trailer packets."
                       .format(self._filename))
                raise Reftek130Exception(msg)
            if len(eh_packets) != 1:
                msg = ("No event header (EH) packets in packet sequence. "
                       "File ({}) might be truncated.".format(self._filename))
                warnings.warn(msg)
            if len(et_packets) != 1:
                msg = ("No event trailer (ET) packets in packet sequence. "
                       "File ({}) might be truncated.".format(self._filename))
                warnings.warn(msg)
            # use either the EH or ET packet, they have the same content (only
            # trigger stop time is not in EH)
            if len(eh_packets):
                eh = EHPacket(eh_packets[0])
            else:
                eh = EHPacket(et_packets[0])
            header = {
                "unit_id": self._data['unit_id'][0],
                "experiment_number": self._data['experiment_number'][0],
                "network": network,
                "station": (eh.station_name +
                            eh.station_name_extension).strip(),
                "location": location, "sampling_rate": eh.sampling_rate,
                "reftek130": eh._to_dict()}
            delta = 1.0 / eh.sampling_rate
            delta_nanoseconds = int(delta * 1e9)
            inds_dt = data['packet_type'] == b"DT"
            data_channels = np.unique(data[inds_dt]['channel_number'])
            for channel_number in data_channels:
                inds = data['channel_number'] == channel_number
                # channel number of EH/ET packets also equals zero (one of the
                # three unused bytes in the extended header of EH/ET packets)
                inds &= data['packet_type'] == b"DT"
                packets = data[inds]

                # split into contiguous blocks, i.e. find gaps. packet sequence
                # was sorted already..
                endtimes = (
                    packets[:-1]["time"] +
                    packets[:-1]["number_of_samples"].astype(np.int64) *
                    delta_nanoseconds)
                # check if next starttime matches seamless to last chunk
                # 1e-3 seconds == 1e6 nanoseconds is the smallest time
                # difference reftek130 format can represent, so anything larger
                # or equal means a gap/overlap.
                time_diffs_milliseconds_abs = np.abs(
                    packets[1:]["time"] - endtimes) / 1000000
                gaps = time_diffs_milliseconds_abs >= 1
                if np.any(gaps):
                    gap_split_indices = np.nonzero(gaps)[0] + 1
                    contiguous = np.array_split(packets, gap_split_indices)
                else:
                    contiguous = [packets]

                for packets_ in contiguous:
                    starttime = packets_[0]['time']

                    if headonly:
                        sample_data = np.array([], dtype=np.int32)
                        npts = packets_["number_of_samples"].sum()
                    else:
                        sample_data = (packets_['payload'][:, :4])
                        sample_data = sample_data.view(np.dtype('>i4')).squeeze()
                        npts = len(sample_data)
                    tr = DiscontinuousTrace(
                        data=sample_data, header=copy.deepcopy(header),
                        times=(packets_['time'] / 10**9).round(3)
                    )
                    tr.stats.npts = packets_['number_of_samples'].sum()
                    tr.stats.actual_npts = npts
                    # channel number is not included in the EH/ET packet
                    # payload, so add it to stats as well..
                    tr.stats.reftek130['channel_number'] = channel_number
                    tr.stats.starttime = UTCDateTime(ns=starttime)
                    """
                    if component codes were explicitly provided, use them
                    together with the stream label
                    if component_codes is not None:
                        tr.stats.channel = (eh.stream_name.strip() +
                                            component_codes[channel_number])
                    # otherwise check if channel code is set for the given
                    # channel (seems to be not the case usually)
                    elif eh.channel_code[channel_number] is not None:
                        tr.stats.channel = eh.channel_code[channel_number]
                    # otherwise fall back to using the stream label together
                    # with the number of the channel in the file (starting with
                    # 0, as Z-1-2 is common use for data streams not oriented
                    # against North)
                    else:
                        msg = ("No channel code specified in the data file "
                               "and no component codes specified. Using "
                               "stream label and number of channel in file as "
                               "channel codes.")
                        warnings.warn(msg)
                        tr.stats.channel = (
                            eh.stream_name.strip() + str(channel_number))
                    """
                    DS = self._data['data_stream_number'][0] + 1
                    if DS != 9:
                        tr.stats.channel = "DS%s-%s" % (DS, channel_number + 1)
                    else:
                        if not include_mp123 and channel_number < 3:
                            continue
                        if not include_mp456 and channel_number > 2:
                            continue
                        tr.stats.channel = "MassPos%s" % (channel_number + 1)
                        # check if endtime of trace is consistent
                    st += tr
        return st
